[Bf-blender-cvs] [ab444a80a28] master: BLI: refactor length parameterization

Jacques Lucke noreply at git.blender.org
Sat Jul 2 21:52:11 CEST 2022


Commit: ab444a80a280e125b3e4d002941504d56f340ced
Author: Jacques Lucke
Date:   Sat Jul 2 21:51:45 2022 +0200
Branches: master
https://developer.blender.org/rBab444a80a280e125b3e4d002941504d56f340ced

BLI: refactor length parameterization

This refactor had two main goals:
* Simplify the sampling code by using an algorithm with fewer special cases.
* Generalize the sampling to support non-sorted samples.

The `SampleSegmentHint` optimization was inspired by `ValueAccessor` from
OpenVDB and improves performance 2x in my test cases.

Differential Revision: https://developer.blender.org/D15348

===================================================================

M	source/blender/blenlib/BLI_length_parameterize.hh
M	source/blender/blenlib/intern/length_parameterize.cc
M	source/blender/blenlib/tests/BLI_length_parameterize_test.cc
M	source/blender/editors/sculpt_paint/curves_sculpt_brush.cc
M	source/blender/geometry/intern/resample_curves.cc

===================================================================

diff --git a/source/blender/blenlib/intern/length_parameterize.cc b/source/blender/blenlib/intern/length_parameterize.cc
index e18b048e96d..06cca281510 100644
--- a/source/blender/blenlib/intern/length_parameterize.cc
+++ b/source/blender/blenlib/intern/length_parameterize.cc
@@ -1,144 +1,58 @@
 /* SPDX-License-Identifier: GPL-2.0-or-later */
 
 #include "BLI_length_parameterize.hh"
+#include "BLI_task.hh"
 
 namespace blender::length_parameterize {
 
-void create_uniform_samples(const Span<float> lengths,
-                            const bool cyclic,
-                            MutableSpan<int> indices,
-                            MutableSpan<float> factors)
+void sample_uniform(const Span<float> lengths,
+                    const bool include_last_point,
+                    MutableSpan<int> r_segment_indices,
+                    MutableSpan<float> r_factors)
 {
-  const int count = indices.size();
+  const int count = r_segment_indices.size();
   BLI_assert(count > 0);
   BLI_assert(lengths.size() >= 1);
   BLI_assert(std::is_sorted(lengths.begin(), lengths.end()));
-  const int segments_num = lengths.size();
-  const int points_num = cyclic ? segments_num : segments_num + 1;
 
-  indices.first() = 0;
-  factors.first() = 0.0f;
   if (count == 1) {
+    r_segment_indices[0] = 0;
+    r_factors[0] = 0.0f;
     return;
   }
-
   const float total_length = lengths.last();
-  if (total_length == 0.0f) {
-    indices.fill(0);
-    factors.fill(0.0f);
-    return;
-  }
-
-  const float step_length = total_length / (count - (cyclic ? 0 : 1));
-  const float step_length_inv = 1.0f / step_length;
-
-  int i_dst = 1;
-  /* Store the length at the previous point in a variable so it can start out at zero
-   * (the lengths array doesn't contain 0 for the first point). */
-  float prev_length = 0.0f;
-  for (const int i_src : IndexRange(points_num - 1)) {
-    const float next_length = lengths[i_src];
-    const float segment_length = next_length - prev_length;
-    if (segment_length == 0.0f) {
-      continue;
-    }
-    /* Add every sample that fits in this segment. */
-    const float segment_length_inv = 1.0f / segment_length;
-    const int segment_samples_num = std::ceil(next_length * step_length_inv - i_dst);
-    indices.slice(i_dst, segment_samples_num).fill(i_src);
-
-    for (const int i : IndexRange(i_dst, s

@@ Diff output truncated at 10240 characters. @@



More information about the Bf-blender-cvs mailing list
diff --git a/source/blender/blenlib/BLI_length_parameterize.hh b/source/blender/blenlib/BLI_length_parameterize.hh
index ec0fabb75b4..ff957d92263 100644
--- a/source/blender/blenlib/BLI_length_parameterize.hh
+++ b/source/blender/blenlib/BLI_length_parameterize.hh
@@ -19,7 +19,7 @@ namespace blender::length_parameterize {
  *
  * \note This is the same as #bke::curves::curve_segment_num.
  */
-inline int lengths_num(const int points_num, const bool cyclic)
+inline int segments_num(const int points_num, const bool cyclic)
 {
   return cyclic ? points_num : points_num - 1;
 }
@@ -30,7 +30,7 @@ inline int lengths_num(const int points_num, const bool cyclic)
 template<typename T>
 void accumulate_lengths(const Span<T> values, const bool cyclic, MutableSpan<float> lengths)
 {
-  BLI_assert(lengths.size() == lengths_num(values.size(), cyclic));
+  BLI_assert(lengths.size() == segments_num(values.size(), cyclic));
   float length = 0.0f;
   for (const int i : IndexRange(values.size() - 1)) {
     length += math::distance(values[i], values[i + 1]);
@@ -42,57 +42,122 @@ void accumulate_lengths(const Span<T> values, const bool cyclic, MutableSpan<flo
 }
 
 template<typename T>
-void linear_interpolation(const Span<T> src,
-                          const Span<int> indices,
-                          const Span<float> factors,
-                          MutableSpan<T> dst)
+inline void linear_interpolation(const Span<T> src,
+                                 const Span<int> indices,
+                                 const Span<float> factors,
+                                 MutableSpan<T> dst)
 {
   BLI_assert(indices.size() == factors.size());
   BLI_assert(indices.size() == dst.size());
-  const int last_src_index = src.index_range().last();
+  const int last_src_index = src.size() - 1;
 
-  int cyclic_sample_count = 0;
-  for (int i = indices.index_range().last(); i > 0; i--) {
-    if (indices[i] != last_src_index) {
-      break;
+  for (const int i : dst.index_range()) {
+    const int prev_index = indices[i];
+    const float factor = factors[i];
+    const bool is_cyclic_case = prev_index == last_src_index;
+    if (is_cyclic_case) {
+      dst[i] = math::interpolate(src.last(), src.first(), factor);
+    }
+    else {
+      dst[i] = math::interpolate(src[prev_index], src[prev_index + 1], factor);
+    }
+  }
+}
+
+/**
+ * Passing this to consecutive calls of #sample_at_length can increase performance.
+ */
+struct SampleSegmentHint {
+  int segment_index = -1;
+  float segment_start;
+  float segment_length_inv;
+};
+
+/**
+ * \param accumulated_segment_lengths: Lengths of individual segments added up.
+ * \param sample_length: The position to sample at.
+ * \param r_segment_index: Returns the index of the segment that #sample_length is in.
+ * \param r_factor: Returns the position within the segment.
+ *
+ * \note #sample_length must not be outside of any segment.
+ */
+inline void sample_at_length(const Span<float> accumulated_segment_lengths,
+                             float sample_length,
+                             int &r_segment_index,
+                             float &r_factor,
+                             SampleSegmentHint *hint = nullptr)
+{
+  /* Use a shorter variable name. */
+  const Span<float> lengths = accumulated_segment_lengths;
+
+  BLI_assert(lengths.size() > 0);
+  BLI_assert(sample_length >= 0.0f);
+  BLI_assert(sample_length <= lengths.last());
+
+  if (hint != nullptr && hint->segment_index >= 0) {
+    const float length_in_segment = sample_length - hint->segment_start;
+    const float factor = length_in_segment * hint->segment_length_inv;
+    if (factor >= 0.0f && factor < 1.0f) {
+      r_segment_index = hint->segment_index;
+      r_factor = factor;
+      return;
     }
-    dst[i] = math::interpolate(src.last(), src.first(), factors[i]);
-    cyclic_sample_count++;
   }
 
-  for (const int i : dst.index_range().drop_back(cyclic_sample_count)) {
-    dst[i] = math::interpolate(src[indices[i]], src[indices[i] + 1], factors[i]);
+  const float total_length = lengths.last();
+  if (sample_length >= total_length) {
+    /* Return the last position on the last segment. */
+    r_segment_index = lengths.size() - 1;
+    r_factor = 1.0f;
+    return;
+  }
+
+  const int prev_point_index = std::upper_bound(lengths.begin(), lengths.end(), sample_length) -
+                               lengths.begin();
+  const float segment_start = prev_point_index == 0 ? 0.0f : lengths[prev_point_index - 1];
+  const float segment_end = lengths[prev_point_index];
+  const float segment_length = segment_end - segment_start;
+  const float segment_length_inv = safe_divide(1.0f, segment_length);
+  const float length_in_segment = sample_length - segment_start;
+  const float factor = length_in_segment * segment_length_inv;
+
+  r_segment_index = prev_point_index;
+  r_factor = factor;
+
+  if (hint != nullptr) {
+    hint->segment_index = r_segment_index;
+    hint->segment_start = segment_start;
+    hint->segment_length_inv = segment_length_inv;
   }
 }
 
 /**
- * Find the given number of points, evenly spaced along the provided length. For non-cyclic
- * sequences, the first point will always be included, and last point will always be included if
- * the #count is greater than zero. For cyclic sequences, the first point will always be included.
+ * Find evenly spaced samples along the lengths.
  *
- * \warning The #count argument must be greater than zero.
+ * \param accumulated_segment_lengths: The accumulated lengths of the original elements being
+ * sampled. Could be calculated by #accumulate_lengths.
+ * \param include_last_point: Generally false for cyclic sequences and true otherwise.
+ * \param r_segment_indices: The index of the previous point at each sample.
+ * \param r_factors: The portion of the length in each segment at each sample.
  */
-void create_uniform_samples(Span<float> lengths,
-                            bool cyclic,
-                            MutableSpan<int> indices,
-                            MutableSpan<float> factors);
+void sample_uniform(Span<float> accumulated_segment_lengths,
+                    bool include_last_point,
+                    MutableSpan<int> r_segment_indices,
+                    MutableSpan<float> r_factors);
 
 /**
  * For each provided sample length, find the segment index and interpolation factor.
  *
  * \param lengths: The accumulated lengths of the original elements being sampled.
  * Could be calculated by #accumulate_lengths.
  * \param sample_lengths: Sampled locations in the #lengths array. Must be sorted and is expected
  * to be within the range of the #lengths values.
  * \param cyclic: Whether the points described by the #lengths input is cyclic. This is likely
  * redundant information theoretically.
  * \param indices: The index of the previous point at each sample.
  * \param factors: The portion of the length in each segment at each sample.
  */
-void create_samples_from_sorted_lengths(Span<float> lengths,
-                                        Span<float> sample_lengths,
-                                        bool cyclic,
-                                        MutableSpan<int> indices,
-                                        MutableSpan<float> factors);
+void sample_at_lengths(Span<float> accumulated_segment_lengths,
+                       Span<float> sample_lengths,
+                       MutableSpan<int> r_segment_indices,
+                       MutableSpan<float> r_factors);
 
 }  // namespace blender::length_parameterize
@@ -42,57 +42,122 @@ void accumulate_lengths(const Span<T> values, const bool cyclic, MutableSpan<flo
  * For each provided sample length, find the segment index and interpolation factor.
  *
- * \param lengths: The accumulated lengths of the original elements being sampled.
- * Could be calculated by #accumulate_lengths.
+ * \param accumulated_segment_lengths: The accumulated lengths of the original elements being
+ * sampled. Could be calculated by #accumulate_lengths.
  * \param sample_lengths: Sampled locations in the #lengths array. Must be sorted and is expected
  * to be within the range of the #lengths values.
@@ -42,57 +42,122 @@ void accumulate_lengths(const Span<T> values, const bool cyclic, MutableSpan<flo
  * \param sample_lengths: Sampled locations in the #lengths array. Must be sorted and is expected
  * to be within the range of the #lengths values.
- * \param cyclic: Whether the points described by the #lengths input is cyclic. This is likely
- * redundant information theoretically.
- * \param indices: The index of the previous point at each sample.
- * \param factors: The portion of the length in each segment at each sample.
+ * \param r_segment_indices: The index of the previous point at each sample.
+ * \param r_factors: The portion of the length in each segment at each sample.
  */